/*
** Module   : MetadataSecurityOps.java
** Abstract : Progress 4GL compatible security interface
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ----------------------------------Description---------------------------------
** 001 OM  20130913 First version, declaring the basic four methods: get/set with and without the
**                  logical database name.
*/

package com.goldencode.p2j.util;

import org.hibernate.ScrollableResults;

import com.goldencode.p2j.directory.*;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.persist.dialect.P2JDialect;

/**
 * Implementation of user security functions of P4GL using the _User metadata from database.
 *
 * From my tests, Progress does not use locks when executing <code>setuserid</code> and
 * <code>userid</code> functions. If a user is exclusive-locking the _user table, another one
 * can get/set userid for the same database. However, attempting to execute a FIND query on the
 * _user meta-table it will be put on hold until the first user releases the lock.
 *
 * TODO:
 * Err 709: "-P and -U startup parameters require _User file records."
 */
public class MetadataSecurityOps implements UserSecurityOps
{
   /** Constant used for accessing the converted _User meta-table. */
   public static final String META_TABLE_USER =  "meta_user";

   /** Constant used for accessing the _userid field of converted _User meta-table. */
   public static final String META_FIELD_USERID =  "meta_userid";

   /** Constant used for accessing the _password field of converted _User meta-table. */
   public static final String META_FIELD_PASSWORD =  "meta_password";

   /**
    * Returns the currently defined default database. This method is only used by the forms of 
    * userid/setuserid methods that do not have the ldbname as explicit parameter.<br/> 
    * The default database *IS* (not yet implemented) automatically connected at startup. If such 
    * database is not defined/connected P4GL will issue an error at compile time:
    * <pre> 
    *     SETUSERID/USERID functions require one connected db or db argument (691).
    * </pre>
    * if the program contains calls to USERID/SETUSERID. <br/>
    * If multiple databases are connected at startup P4GL issues the same compiling error.
    * 
    * @return  the name of the default database if any, or empty string otherwise.
    */
   private String getDefaultDatabase() 
   {
      Directory dir = DirectoryManager.getInstance();
      // explicitly configured in /server/default/runtime/default/
      return dir.getString(Directory.ID_RELATIVE, "defaultDatabase", "");
   }
   
   /**
    * Returns the userid associated with the current connected database. If no database is
    * connected or there are at least two databases connected empty string is returned.
    *
    * @return   The current user context's userid or the empty string in the case of any problem.
    */
   public String getUserId()
   {
      // this 4GL functions tends (?) to return the user id of the "default" database
      // the "default" database is the one connected at compile time
      return getUserIdFromDB(getDefaultDatabase());
   }

   /**
    * Returns the userid associated with the current user of the given logical database.
    *
    * @param    dbname
    *           Logical database name.
    *
    * @return   The current user context's userid or the empty string in the
    *           case of any problem.
    */
   public String getUserIdFromDB(String dbname)
   {
      // null/empty validation
      if (dbname == null || dbname.isEmpty())
      {
         // at runtime, we get an empty string (not unknown, not error)
         return "";
      }

      // validation: dbname MUST be the logical name of a connected database
      character ldbName = ConnectionManager.ldbName(dbname);
      if (ldbName.isUnknown())
      {
         // the 'dbname' is not a valid connected database logical name nor an alias
         // at runtime, we get an empty string (not unknown, not error)
         return "";
      }
      if (!dbname.equalsIgnoreCase(ldbName.toStringMessage()))
      {
         // 'dbname' is probably an alias, not a true logical name
         // at runtime, we get an empty string (not unknown, not error)
         return "";
      }

      // use the cached user id from ConnectionManager
      return ConnectionManager.getCurrentUserid(dbname);
   }

   /**
    * Authenticates the user for specified DB connection. Checks if the user login account match
    * the corresponding record of the _User table of the database. If there are multiple database
    * connections or none is connected, an error is issued.
    *
    * @param    userid
    *           The name of the user to set as UserID.
    * @param    password
    *           The users password.
    *
    * @return   The <code>true</code> in case of valid user match <code>false</code> otherwise.
    */
   public boolean setUserId(String userid, String password)
   {
      return setUserId(userid, password, getDefaultDatabase());
   }

   /**
    * Authenticates the user for specified DB connection. Checks if the user login account match
    * the corresponding record of the _User table of the database.  If the specified database
    * is not connected, an error is issued.
    *
    *
    * @param    userid
    *           The name of the user to set as UserID.
    * @param    password
    *           The users password.
    * @param    dbname
    *           Logical database name.
    *
    * @return   The <code>true</code> in case of valid user match <code>false</code> otherwise.
    */
   public boolean setUserId(String userid, String password, String dbname)
   {
      // validation: dbname MUST be the logical name of a connected database
      character ldbName = ConnectionManager.ldbName(dbname);
      if (dbname == null      ||
          dbname.isEmpty()    ||
          ldbName.isUnknown() ||
          !dbname.equalsIgnoreCase(ldbName.toStringMessage()))
      {
         // the 'dbname' is not a valid connected database logical name nor an alias
         ErrorManager.recordOrShowError(
               1072,
                "SETUSERID of database " + dbname +
                " requires that database to be connected", false);
         // at runtime, we get a false value (not unknown, not error)
         return false;
      }

      String encPassword = SecurityOps.encode(password).toStringMessage();
      Persistence persistence = ConnectionManager.getPersistence(dbname);

      P2JDialect dialect = (P2JDialect) persistence.getDialect();
      String pccUserid = dialect.getProcessedCharacterColumnName(META_FIELD_USERID, true);
      String pccPassword = dialect.getProcessedCharacterColumnName(META_FIELD_PASSWORD, false);

      /* use the dialect or something to execute the query equivalent to following: */
      String query =  "SELECT COUNT(*) FROM " + META_TABLE_USER + " " +
                           "WHERE " + pccUserid + " = ? AND " + pccPassword + " = ? ";
      try
      {
         ScrollableResults rs = persistence.executeSQLQuery(query,
                                          new String[] { userid.toUpperCase(), encPassword });
         if (rs == null || !rs.next())
         {
            // this should not happen anyway
            return false;
         }
         
         long count = ((java.math.BigInteger) rs.get(0)).longValue();
         if (count != 1)
         {
            // no match (if zero), if 2 or more then the _user table is broken
            return false;
         }
      }
      catch (PersistenceException e)
      {
         System.err.println("setUserId failed !!!!!!!!!!");
         return false;
      }
      
      // if we reached this point then the user/password combination is OK, set the new userid
      ConnectionManager.setCurrentUserid(dbname, userid);
      
      return true;
   }


   /**
    * Check the access level for a database. This will query the _User table and return the value.
    * 
    * @param   ldbname
    *          The logical database name to query. 
    * 
    * @return  The access level to specified database. 
    */
   @Override
   public int getAuthLevel(String ldbname)
   {
      Persistence persistence = ConnectionManager.getPersistence(ldbname);

      /* use the dialect or something to execute the query equivalent to following: */
      String query =  "SELECT COUNT(*) FROM " + META_TABLE_USER;
      try
      {
         ScrollableResults rs = persistence.executeSQLQuery(query);
         if (rs != null && rs.next())
         {
            long count = ((java.math.BigInteger) rs.get(0)).longValue();
            return count == 0 ? SecurityOps.FREE_ACCESS : SecurityOps.BLANK_ALLOWED;
         }
         
      }
      catch (PersistenceException e)
      {
         System.err.println("getAuthLevel failed !!!!!!!!!!");
      }
      
      return SecurityOps.FREE_ACCESS;
   }
}
