/*
** Module   : SecurityOps.java
** Abstract : Progress 4GL compatible security interface
**
** Copyright (c) 2006-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- -----------------------------Description-----------------------------
** 001 GES 20060331   @25287 First version implementing the USERID builtin
**                           function.
** 002 GES 20060331   @25294 Added ENCODE support (not Progress 4GL 
**                           compatible).
** 003 NVS 20060418   @25585 Added static convenience method to be called
**                           from the converted code: changePassword().
** 004 GES 20060831   @29063 Make sure encode output only includes chars
**                           from the set A-Z and a-z using a simple
**                           table driven base 52 approach.
** 005 GES 20101005          Added stub (non-functional) methods for the
**                           USERID(dbname) variant.
** 006 CA  20130129          Added stubs for misc security-related builtins.
**                           Implemented base64-encode/decode.
** 007 EVL 20130225          Adding conversion support for SETUSERID built-in.
** 008 GES 20130523          Minor fixes for the base64-encode/decode.
** 009 OM  20130917          Added _User metadata hadling (configured in directory)
*/

package com.goldencode.p2j.util;

import java.security.*;
import java.util.logging.*;

import com.goldencode.p2j.directory.*;
import com.goldencode.p2j.security.SecurityManager;

import org.apache.commons.codec.binary.Base64;

/**
 * Provides static utility methods for Progress 4GL compatible security
 * functions.  At this time, the following Progress 4GL functions are handled:
 * <p>
 * <pre>
 * Progress Function              Java Equivalent
 * ---------------------------    ---------------------------
 * USERID                         {@link #getUserId}
 * USERID(dbname)                 {@link #getUserIdFromDB}
 * ENCODE                         {@link #encode}
 * BASE64-ENCODE                  {@link #base64Encode}
 * BASE64-DECODE                  {@link #base64Decode}
 * ENCRYPT                        {@link #encrypt}
 * GENERATE-PBE-KEY               {@link #generatePBEKey}
 * GENERATE-PBE-SALT              {@link #generatePBESalt}
 * SHA1-DIGEST                    {@link #sha1Digest}
 * MDG-DIGEST                     {@link #md5Digest}
 * </pre>
 * <p>
 *
 * @author    GES
 */
public class SecurityOps
{
   /** _User table of the database is empty, no access privilege needed */
   public static final int FREE_ACCESS = 0;
   
   /** Database has userids stored in _user table, but do not enforce them. */
   public static final int BLANK_ALLOWED = 1;
   
   /** Only with valid user/password are allowed to access a database. */
   public static final int STRICT = 2;

   /** The default classname for _User security handler */
   private static final String DEFAULT_SECURITY_OPS_HANDLER =
         "com.goldencode.p2j.util.SecurityManagerSecurityOps";

   /** Logger */
   private static final Logger LOG = LogHelper.getLogger(SecurityOps.class);

   /** The current _User security handler for emulation of USERID and SETUSERID P4GL functions */
   private static UserSecurityOps userSecurityOps = instantiateUserHandler();

   /**
    * Instantiates and returns a SecurityOps handler for emulation of USERID and SETUSERID P4GL
    * functions as configured in the directory.
    *
    * At this moment, only two of implementation are available: SecurityManagerSecurityOps and
    * DatabaseTableSecurityOps. If the directory does not contain an entry for
    * 'userSecurityOpsClass' then the default SecurityManagerSecurityOps will be used.
    *
    * @return  an instance of SecurityOps handler as configured in directory
    */
   private static UserSecurityOps instantiateUserHandler()
   {
      Directory dir = DirectoryManager.getInstance();
      String handlerName = dir.getString(Directory.ID_RELATIVE,
                                         "userSecurityOpsClass",
                                         DEFAULT_SECURITY_OPS_HANDLER);

      // use the class-loader to get the class into memory
      Class<?> cl = null;
      try
      {
         cl = Class.forName(handlerName);
      }
      catch (Exception exc)
      {
         if (LOG.isLoggable(Level.WARNING))
         {
            LOG.log(Level.WARNING,
                    "Can't load user security handler '" + handlerName + "'. Loading default.",
                    exc);
         }
         return getDefaultUserHandler();
      }

      // attempt to instantiate an object
      Object o = null;
      try
      {
         o = cl.newInstance();
      }
      catch (Exception exc)
      {
         if (LOG.isLoggable(Level.WARNING))
         {
            LOG.log(Level.WARNING,
                    "User security handler '" + handlerName
                          + "'cannot be instantiated. Loading default.",
                    exc);
         }
         return getDefaultUserHandler();
      }

      // make sure the object is of the desired type
      if (!(o instanceof UserSecurityOps))
      {
         if (LOG.isLoggable(Level.WARNING))
         {
            LOG.log(Level.WARNING,
                    "Class '" + handlerName
                          + "'cannot be used as security handler. Loading default.");
         }
         return getDefaultUserHandler();
      }

      return (UserSecurityOps) o;
   }

   /**
    * Instantiate the default _User handler and returns it. This method must not be called
    * multiple times as only one _User handler should exist.
    *
    * @return  An instance of UserSecurityOps
    */
   private static UserSecurityOps getDefaultUserHandler()
   {
      return new SecurityManagerSecurityOps();
   }

   /** Table of all alphabetic characters used to limit the encode output. */
   private static final String ALPHAS = "ABCDEFGHIJKLMNOPQRSTUVWXYZ" +
                                        "abcdefghijklmnopqrstuvwxyz";

   /**
    * Returns the userid associated with the current user.
    * <p>
    * Note that in Progress 4GL, this takes an optional string parameter with
    * the name of the logical database in which the userid is set.  There is
    * no corresponding database level authentication concept in P2J so this
    * parameter is not supported.
    *
    * @return   The current user context's userid or the empty string in the
    *           case of any problem.
    */
   public static character getUserId() 
   {
      return new character(userSecurityOps.getUserId());
   }
   
   /**
    * Returns the userid associated with the current user of the given logical database.
    * <p>
    * This will call the current configured implementation of UserSecurityOps. If the database
    * _User metadata is enabled then the authenticated user is returned, otherwise the default
    * SecurityManager user is returned.
    *
    * @param    dbname
    *           Logical database name.
    *
    * @return   The current user context's userid or the empty string in the
    *           case of any problem.
    */
   public static character getUserIdFromDB(character dbname)
   {
      // call the method with unwrapped-argument signature:
      return getUserIdFromDB(dbname.toStringMessage());
   }
   
   /**
    * Returns the userid associated with the current user of the given logical database.
    * <p>
    * This will call the current configured implementation of UserSecurityOps. If the database
    * _User metadata is enabled then the authenticated user is returned, otherwise the default
    * SecurityManager user is returned.
    *
    * @param    dbname
    *           Logical database name.
    *
    * @return   The current user context's userid or the empty string in the
    *           case of any problem.
    */
   public static character getUserIdFromDB(String dbname)
   {
      return new character(userSecurityOps.getUserIdFromDB(dbname));
   }
   
   /**
    * Authenticates the user for specified DB connection. Checks if the user login account match
    * the corresponding record of the _User table of the database.
    *
    * @param    userid
    *           The name of the user to set as UserID.
    * @param    password
    *           The users password.
    * @param    dbname
    *           Logical database name.
    *
    * @return   The <code>true</code> in case of valid user match <code>false</code> otherwise.
    */
   public static logical setUserId(String userid, character password, character dbname) 
   {
      return new logical(userSecurityOps.setUserId(userid,
                                                   password.toStringMessage(),
                                                   dbname.toStringMessage()));
   }
   
   /**
    * Authenticates the user for specified DB connection. Checks if the user login account match
    * the corresponding record of the _User table of the database.
    *
    * @param    userid
    *           The name of the user to set as UserID.
    * @param    password
    *           The users password.
    * @param    dbname
    *           Logical database name.
    *
    * @return   The <code>true</code> in case of valid user match <code>false</code> otherwise.
    */
   public static logical setUserId(character userid, String password, character dbname) 
   {
      return new logical(userSecurityOps.setUserId(userid.toStringMessage(),
                                                   password,
                                                   dbname.toStringMessage()));
   }
   
   /**
    * Authenticates the user for specified DB connection. Checks if the user login account match
    * the corresponding record of the _User table of the database.
    *
    * @param    userid
    *           The name of the user to set as UserID.
    * @param    password
    *           The users password.
    * @param    dbname
    *           Logical database name.
    *
    * @return   The <code>true</code> in case of valid user match <code>false</code> otherwise.
    */
   public static logical setUserId(String userid, String password, character dbname) 
   {
      return new logical(userSecurityOps.setUserId(userid, password, dbname.toStringMessage()));
   }

   /**
    * Authenticates the user for specified DB connection. Checks if the user login account match
    * the corresponding record of the _User table of the database.
    *
    * @param    userid
    *           The name of the user to set as UserID.
    * @param    password
    *           The users password.
    * @param    dbname
    *           Logical database name.
    *
    * @return   The <code>true</code> in case of valid user match <code>false</code> otherwise.
    */
   public static logical setUserId(character userid, character password, String dbname)
   {
      return new logical(userSecurityOps.setUserId(userid.toStringMessage(),
                                                   password.toStringMessage(),
                                                   dbname));
   }
   
   /**
    * Authenticates the user for specified DB connection. Checks if the user login account match
    * the corresponding record of the _User table of the database.
    *
    * @param    userid
    *           The name of the user to set as UserID.
    * @param    password
    *           The users password.
    * @param    dbname
    *           Logical database name.
    *
    * @return   The <code>true</code> in case of valid user match <code>false</code> otherwise.
    */
   public static logical setUserId(String userid, character password, String dbname) 
   {
      return new logical(userSecurityOps.setUserId(userid, password.toStringMessage(), dbname));
   }
   
   /**
    * Authenticates the user for specified DB connection. Checks if the user login account match
    * the corresponding record of the _User table of the database.
    *
    * @param    userid
    *           The name of the user to set as UserID.
    * @param    password
    *           The users password.
    * @param    dbname
    *           Logical database name.
    *
    * @return   The <code>true</code> in case of valid user match <code>false</code> otherwise.
    */
   public static logical setUserId(character userid, String password, String dbname) 
   {
      return new logical(userSecurityOps.setUserId(userid.toStringMessage(), password, dbname));
   }
   
   /**
    * Authenticates the user for specified DB connection. Checks if the user login account match
    * the corresponding record of the _User table of the database.
    *
    * @param    userid
    *           The name of the user to set as UserID.
    * @param    password
    *           The users password.
    * @param    dbname
    *           Logical database name.
    *
    * @return   The <code>true</code> in case of valid user match <code>false</code> otherwise.
    */
   public static logical setUserId(String userid, String password, String dbname) 
   {
      return new logical(userSecurityOps.setUserId(userid, password, dbname));
   }
   
   /**
    * Authenticates the user for specified DB connection. Checks if the user login account match
    * the corresponding record of the _User table of the database.
    *
    * @param    userid
    *           The name of the user to set as UserID.
    * @param    password
    *           The users password.
    *
    * @return   The <code>true</code> in case of valid user match <code>false</code> otherwise.
    */
   public static logical setUserId(String userid, character password) 
   {
      return new logical(userSecurityOps.setUserId(userid, password.toStringMessage()));
   }
   
   /**
    * Authenticates the user for specified DB connection. Checks if the user login account match
    * the corresponding record of the _User table of the database.
    *
    * @param    userid
    *           The name of the user to set as UserID.
    * @param    password
    *           The users password.
    *
    * @return   The <code>true</code> in case of valid user match <code>false</code> otherwise.
    */
   public static logical setUserId(character userid, String password) 
   {
      return new logical(userSecurityOps.setUserId(userid.toStringMessage(), password));
   }
   
   /**
    * Authenticates the user for specified DB connection. Checks if the user login account match
    * the corresponding record of the _User table of the database.
    *
    * @param    userid
    *           The name of the user to set as UserID.
    * @param    password
    *           The users password.
    *
    * @return   The <code>true</code> in case of valid user match <code>false</code> otherwise.
    */
   public static logical setUserId(String userid, String password) 
   {
      return new logical(userSecurityOps.setUserId(userid, password));
   }
   
   /**
    * Authenticates the user for specified DB connection. Checks if the user login account match
    * the corresponding record of the _User table of the database.
    *
    * @param    userid
    *           The name of the user to set as UserID.
    * @param    password
    *           The users password.
    *
    * @return   The <code>true</code> in case of valid user match <code>false</code> otherwise.
    */
   public static logical setUserId(character userid, character password) 
   {
      return new logical(userSecurityOps.setUserId(userid.toStringMessage(),
                                                   password.toStringMessage()));
   }
   
   /**
    * Authenticates the user for specified DB connection. Checks if the user login account match
    * the corresponding record of the _User table of the database.
    *
    * @param    userid
    *           The name of the user to set as UserID.
    * @param    password
    *           The users password.
    * @param    dbname
    *           Logical database name.
    *
    * @return   The <code>true</code> in case of valid user match <code>false</code> otherwise.
    */
   public static logical setUserId(character userid, character password, character dbname)
   {
      return new logical(userSecurityOps.setUserId(userid.toStringMessage(),
                                                   password.toStringMessage(),
                                                   dbname.toStringMessage()));
   }
   
   /**
    * Check the access level for a database. The response depend on UserSecurityOps implementation
    * on which this method will forward the message.
    * 
    * @param   ldbname
    *          The logical database name to query. 
    * 
    * @return  The access level to specified database. 
    */
   public static int getAuthLevel(String ldbname)
   {
      return userSecurityOps.getAuthLevel(ldbname);
   }
   
   /**
    * Performs a password change procedure for the currently logged on user.
    * <p>
    * Application has no control over the procedure itself. 
    *
    * @return   failure/success indication
    *           <code>true></code> means 
    */
   public static logical changePassword() 
   {
      logical result = new logical(false);
      SecurityManager sm = SecurityManager.getInstance();
      
      if (sm != null)
      {
         result = new logical(sm.changePassword());
      }
      
      return result;
   }

   /**
    * Hashes an arbitrary input string into a 16-byte output string.  This
    * conversion is one-way (the original data can not be recovered) and the
    * output is always a string of 16-bytes no matter the length of the input
    * string.
    * <p>
    * <b>WARNING: this implementation is NOT compatible with the Progress
    * 4GL algorithm in any way. Any programs relying on comparisons between
    * data encoded by the Progress 4GL ENCODE and the return value from this
    * method will NOT compare as expected.</b>
    *
    * @param    data
    *           The data to be hashed.  Must not be <code>null</code>.
    *
    * @return   The hashed 16-byte string or the <code>unknown value</code>
    *           if the input is <code>unknown value</code>.
    */
   public static character encode(character data)
   {
      if (data.isUnknown())
      {
         return new character();
      }
      
      return encode(data.getValue());
   }
   
   /**
    * Hashes an arbitrary input string into a 16-byte output string.  This
    * conversion is one-way (the original data can not be recovered) and the
    * output is always a string of 16-bytes no matter the length of the input
    * string.
    * <p>
    * <b>WARNING: this implementation is NOT compatible with the Progress
    * 4GL algorithm in any way. Any programs relying on comparisons between
    * data encoded by the Progress 4GL ENCODE and the return value from this
    * method will NOT compare as expected.</b>
    * <p>
    * The 16-bytes of output will only include characters in the range
    * <code>A-Z</code> and <code>a-z</code>.
    *
    * @param    data
    *           The data to be hashed.  Must not be <code>null</code>.
    *
    * @return   The hashed 16-byte string.
    */
   public static character encode(String data)
   {
      return new character(_encode(data));
   }

   /**
    * Hashes an arbitrary input string into a 16-byte output string.  This conversion is one-way
    * (the original data can not be recovered) and the output is always a string of 16-bytes no
    * matter the length of the input string.
    * <p>
    * <b>WARNING: this implementation is NOT compatible with the Progress 4GL algorithm in any way.
    * Any programs relying on comparisons between data encoded by the Progress 4GL ENCODE and the
    * return value from this method will NOT compare as expected.</b>
    * <p>
    * The 16-bytes of output will only include characters in the range <code>A-Z</code> and
    * <code>a-z</code>.
    *
    * @param    data
    *           The data to be hashed. Must not be <code>null</code>.
    *
    * @return   The hashed 16-byte string.
    */
   public static String _encode(String data)
   {
      StringBuffer sb = new StringBuffer();
      byte[] hashed = null;
      
      try
      {
         MessageDigest md = MessageDigest.getInstance("MD5");
         hashed = md.digest(data.getBytes());
         
         for (int i = 0; i < hashed.length; i++)
         {
            // this is not a fully even distribution but it is close
            // enough for an interrim solution
            int base52 = ((0x000000FF & hashed[i]) % 52); 
            sb.append(ALPHAS.charAt(base52));
         }
      }
      catch (NoSuchAlgorithmException nsa)
      {
         // we will fail below
      }
      
      if (hashed == null || hashed.length != 16)
      {
         throw new RuntimeException("Unexpected hashing result for input '" +
                                    data + "' resulting in output '" +
                                    hashed + "'.");
      }
      
      return sb.toString();
   }   
   
   /**
    * Converts binary data into a Base64 encoded text.
    *
    * @param    data
    *           The data to be encoded.
    *
    * @return   The encoded string.  <code>unknown</code> will be returned if the given data is
    *           <code>unknown</code> OR if the given data is initialized but the runtime has no
    *           knowledge of the buffer's size.
    */
   public static longchar base64Encode(BinaryData data)
   {
      if (data == null)
      {
         return new longchar();
      }
      
      // TODO: negative memptr sizes support
      // larger negative reported sizes for a memptr can cause this error:
      // Unable to allocate memory for result from BASE64-ENCODE function (12118)
      // a size of -1 can cause a 0-length string as a result
      // neither of these results are properly supported here
      
      return base64Encode(data.getByteArray());
   }

   /**
    * Converts binary data into a Base64 encoded text.
    *
    * @param    data
    *           The data to be encoded.
    *
    * @return   The encoded string.  <code>unknown</code> will be returned if the given data is
    *           <code>null</code>.
    */
   public static longchar base64Encode(byte[] data)
   {
      if (data == null)
      {
         return new longchar();
      }

      // TODO: test the results to see if empty string and other data inputs get the
      // proper compatible result
      return new longchar(new String(Base64.encodeBase64(data)));
   }

   /**
    * Decodes a base64 encoded string into a binary value.
    *
    * @param    data
    *           The string to be decoded.
    *
    * @return   a memptr with the decoded data.
    */
   public static memptr base64Decode(Text data)
   {
      //TODO - review
      if (data.isUnknown()) 
      {
         return new memptr();
      }

      return base64Decode(data.getValue());
   }

   /**
    * Converts a base64 encoded string into a binary value and returns a 
    * memptr with the decoded value.
    *
    * @param    data
    *           The string to be decoded.
    *
    * @return   a memptr with the decoded data.
    */
   public static memptr base64Decode(String data)
   {
      byte[] decoded = Base64.decodeBase64(data.getBytes());
      return new memptr(decoded);
   }
   
   /**
    * Encrypts the source and returns a {@link memptr} instance containing 
    * the encrypted data.
    * 
    * @param    data
    *           The data to be encrypted.
    *           
    * @return   See above.
    */
   public static memptr encrypt(String data)
   {
      return encrypt(data, null);
   }
   
   /**
    * Encrypts the source using the given key and returns a {@link memptr}
    * instance containing the encrypted data.
    * 
    * @param    data
    *           The data to be encrypted.
    * @param    key
    *           The encryption key.
    *           
    * @return   See above.
    */
   public static memptr encrypt(String data, raw key)
   {
      return encrypt(data, key, null);
   }
   
   /**
    * Encrypts the source using the given key and initialization vector, and 
    * returns a {@link memptr} instance containing the encrypted data.
    * 
    * @param    data
    *           The data to be encrypted.
    * @param    key
    *           The encryption key.
    * @param    ivValue
    *           An initialization vector to be used with the given key.
    *           
    * @return   See above.
    */
   public static memptr encrypt(String data, raw key, raw ivValue)
   {
      return encrypt(data, key, ivValue, (String) null);
   }
   
   /**
    * Encrypts the source using the given key, initialization vector and 
    * symmetric encryption algorithm, and returns a {@link memptr} instance 
    * containing the encrypted data.
    * 
    * @param    data
    *           The data to be encrypted.
    * @param    key
    *           The encryption key.
    * @param    ivValue
    *           An initialization vector to be used with the given key.
    * @param    algorithm
    *           The encryption algorithm's name.
    *                      
    * @return   See above.
    */
   public static memptr encrypt(String data, raw key, raw ivValue, String algorithm)
   {
      return encrypt(data, key, ivValue, new character(algorithm));
   }
   
   /**
    * Encrypts the source using the given key, initialization vector and 
    * symmetric encryption algorithm, and returns a {@link memptr} instance 
    * containing the encrypted data.
    * 
    * @param    data
    *           The data to be encrypted.
    * @param    key
    *           The encryption key.
    * @param    ivValue
    *           An initialization vector to be used with the given key.
    * @param    algorithm
    *           The encryption algorithm's name.
    *                      
    * @return   See above.
    */
   public static memptr encrypt(String data, raw key, raw ivValue, character algorithm)
   {
      return encrypt(new character(data), key, ivValue, algorithm);
   }
   
   /**
    * Encrypts the source and returns a {@link memptr} instance containing 
    * the encrypted data.
    * 
    * @param    data
    *           The data to be encrypted.
    *           
    * @return   See above.
    */
   public static memptr encrypt(Text data)
   {
      return encrypt(data, null);
   }
   
   /**
    * Encrypts the source using the given key and returns a {@link memptr}
    * instance containing the encrypted data.
    * 
    * @param    data
    *           The data to be encrypted.
    * @param    key
    *           The encryption key.
    *           
    * @return   See above.
    */
   public static memptr encrypt(Text data, raw key)
   {
      return encrypt(data, key, null);
   }
   
   /**
    * Encrypts the source using the given key and initialization vector, and 
    * returns a {@link memptr} instance containing the encrypted data.
    * 
    * @param    data
    *           The data to be encrypted.
    * @param    key
    *           The encryption key.
    * @param    ivValue
    *           An initialization vector to be used with the given key.
    *           
    * @return   See above.
    */
   public static memptr encrypt(Text data, raw key, raw ivValue)
   {
      return encrypt(data, key, ivValue, (String) null);
   }
   
   /**
    * Encrypts the source using the given key, initialization vector and 
    * symmetric encryption algorithm, and returns a {@link memptr} instance 
    * containing the encrypted data.
    * 
    * @param    data
    *           The data to be encrypted.
    * @param    key
    *           The encryption key.
    * @param    ivValue
    *           An initialization vector to be used with the given key.
    * @param    algorithm
    *           The encryption algorithm's name.
    *                      
    * @return   See above.
    */
   public static memptr encrypt(Text data, raw key, raw ivValue, String algorithm)
   {
      return encrypt(data, key, ivValue, new character(algorithm));
   }

   /**
    * Encrypts the source using the given key, initialization vector and 
    * symmetric encryption algorithm, and returns a {@link memptr} instance 
    * containing the encrypted data.
    * 
    * @param    data
    *           The data to be encrypted.
    * @param    key
    *           The encryption key.
    * @param    ivValue
    *           An initialization vector to be used with the given key.
    * @param    algorithm
    *           The encryption algorithm's name.
    *                      
    * @return   See above.
    */
   public static memptr encrypt(Text data, raw key, raw ivValue, character algorithm)
   {
      // TODO: IMPLEMENT THIS
      return null;
   }
   
   /**
    * Generates a password-based encryption key, based on the PKCS#5/RFC 2898
    * standard, and returns the key as a {@link raw} instance.
    * 
    * @param    password
    *           The password to use in generating the encryption key.
    *           
    * @return   See above.
    */
   public static raw generatePBEKey(String password)
   {
      return generatePBEKey(new character(password), null);
   }
   
   /**
    * Generates a password-based encryption key, based on the PKCS#5/RFC 2898
    * standard, and returns the key as a {@link raw} instance.
    * 
    * @param    password
    *           The password to use in generating the encryption key.
    * @param    salt
    *           a random series of 8 bytes to use as salt value in generating 
    *           the encryption key.
    *           
    * @return   See above.
    */
   public static raw generatePBEKey(String password, raw salt)
   {
      return generatePBEKey(new character(password), salt);
   }
   
   /**
    * Generates a password-based encryption key, based on the PKCS#5/RFC 2898
    * standard, and returns the key as a {@link raw} instance.
    * 
    * @param    password
    *           The password to use in generating the encryption key.
    *           
    * @return   See above.
    */
   public static raw generatePBEKey(Text password)
   {
      return generatePBEKey(password, null);
   }
   
   /**
    * Generates a password-based encryption key, based on the PKCS#5/RFC 2898
    * standard, and returns the key as a {@link raw} instance.
    * 
    * @param    password
    *           The password to use in generating the encryption key.
    * @param    salt
    *           a random series of 8 bytes to use as salt value in generating 
    *           the encryption key.
    *           
    * @return   See above.
    */
   public static raw generatePBEKey(Text password, raw salt)
   {
      // TODO: IMPLEMENT THIS
      return null;
   }
   
   /**
    * Generates a password-based encryption key, based on the PKCS#5/RFC 2898
    * standard, and returns the key as a {@link raw} instance.
    * 
    * @param    password
    *           The password to use in generating the encryption key.
    *           
    * @return   See above.
    */
   public static raw generatePBEKey(BinaryData password)
   {
      return generatePBEKey(password, null);
   }
   
   /**
    * Generates a password-based encryption key, based on the PKCS#5/RFC 2898
    * standard, and returns the key as a {@link raw} instance.
    * 
    * @param    password
    *           The password to use in generating the encryption key.
    * @param    salt
    *           a random series of 8 bytes to use as salt value in generating 
    *           the encryption key.
    *           
    * @return   See above.
    */
   public static raw generatePBEKey(BinaryData password, raw salt)
   {
      // TODO: IMPLEMENT THIS
      return null;
   }
   
   /**
    * Generates a random salt value (a series of 8 bytes) to use in generating
    * an encryption key, and returns the salt value as a {@link raw} instance.
    * 
    * @return   See above.
    */
   public static raw generatePBESalt()
   {
      // TODO: IMPLEMENT THIS
      return new raw();
   }
   
   /**
    * Hashes the specified data using the SHA-1 algorithm, and returns a 
    * 20-byte binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    *           
    * @return   The hash value.
    */
   public static raw sha1Digest(String data)
   {
      return sha1Digest(data, (String) null);
   }
   
   /**
    * Hashes the specified data using the SHA-1 algorithm, and returns a 
    * 20-byte binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    * @param    key
    *           An optional key value to use in the hash operation.
    *           
    * @return   The hash value.
    */
   public static raw sha1Digest(String data, String key)
   {
      return sha1Digest(data, new character(key));
   }
   
   /**
    * Hashes the specified data using the SHA-1 algorithm, and returns a 
    * 20-byte binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    * @param    key
    *           An optional key value to use in the hash operation.
    *           
    * @return   The hash value.
    */
   public static raw sha1Digest(String data, Text key)
   {
      return sha1Digest(new character(data), key);
   }

   /**
    * Hashes the specified data using the SHA-1 algorithm, and returns a 
    * 20-byte binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    * @param    key
    *           An optional key value to use in the hash operation.
    *           
    * @return   The hash value.
    */
   public static raw sha1Digest(String data, BinaryData key)
   {
      return sha1Digest(new character(data), key);
   }
   
   /**
    * Hashes the specified data using the SHA-1 algorithm, and returns a 
    * 20-byte binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    *           
    * @return   The hash value.
    */
   public static raw sha1Digest(Text data)
   {
      return sha1Digest(data, (String) null);
   }
   
   /**
    * Hashes the specified data using the SHA-1 algorithm, and returns a 
    * 20-byte binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    * @param    key
    *           An optional key value to use in the hash operation.
    *           
    * @return   The hash value.
    */
   public static raw sha1Digest(Text data, String key)
   {
      return sha1Digest(data, new character(key));
   }
   
   /**
    * Hashes the specified data using the SHA-1 algorithm, and returns a 
    * 20-byte binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    * @param    key
    *           An optional key value to use in the hash operation.
    *           
    * @return   The hash value.
    */
   public static raw sha1Digest(Text data, Text key)
   {
      // TODO: IMPLEMENT THIS
      return null;
   }

   /**
    * Hashes the specified data using the SHA-1 algorithm, and returns a 
    * 20-byte binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    * @param    key
    *           An optional key value to use in the hash operation.
    *           
    * @return   The hash value.
    */
   public static raw sha1Digest(Text data, BinaryData key)
   {
      // TODO: IMPLEMENT THIS
      return null;
   }
   
   /**
    * Hashes the specified data using the SHA-1 algorithm, and returns a 
    * 20-byte binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    *           
    * @return   The hash value.
    */
   public static raw sha1Digest(BinaryData data)
   {
      return sha1Digest(data, (String) null);
   }
   
   /**
    * Hashes the specified data using the SHA-1 algorithm, and returns a 
    * 20-byte binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    * @param    key
    *           An optional key value to use in the hash operation.
    *           
    * @return   The hash value.
    */
   public static raw sha1Digest(BinaryData data, String key)
   {
      return sha1Digest(data, new character(key));
   }
   
   /**
    * Hashes the specified data using the SHA-1 algorithm, and returns a 
    * 20-byte binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    * @param    key
    *           An optional key value to use in the hash operation.
    *           
    * @return   The hash value.
    */
   public static raw sha1Digest(BinaryData data, Text key)
   {
      // TODO: IMPLEMENT THIS
      return null;
   }

   /**
    * Hashes the specified data using the SHA-1 algorithm, and returns a 
    * 20-byte binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    * @param    key
    *           An optional key value to use in the hash operation.
    *           
    * @return   The hash value.
    */
   public static raw sha1Digest(BinaryData data, BinaryData key)
   {
      // TODO: IMPLEMENT THIS
      return null;
   }

   /**
    * Hashes the specified data using RSA MD5 algorithm, and returns a 16-byte 
    * binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    *           
    * @return   The hash value.
    */
   public static raw md5Digest(String data)
   {
      return md5Digest(data, (String) null);
   }
   
   /**
    * Hashes the specified data using RSA MD5 algorithm, and returns a 16-byte 
    * binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    * @param    key
    *           An optional key value to use in the hash operation.
    *           
    * @return   The hash value.
    */
   public static raw md5Digest(String data, String key)
   {
      return md5Digest(data, new character(key));
   }
   
   /**
    * Hashes the specified data using RSA MD5 algorithm, and returns a 16-byte 
    * binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    * @param    key
    *           An optional key value to use in the hash operation.
    *           
    * @return   The hash value.
    */
   public static raw md5Digest(String data, Text key)
   {
      return md5Digest(new character(data), key);
   }

   /**
    * Hashes the specified data using RSA MD5 algorithm, and returns a 16-byte 
    * binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    * @param    key
    *           An optional key value to use in the hash operation.
    *           
    * @return   The hash value.
    */
   public static raw md5Digest(String data, BinaryData key)
   {
      return md5Digest(new character(data), key);
   }
   
   /**
    * Hashes the specified data using RSA MD5 algorithm, and returns a 16-byte 
    * binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    *           
    * @return   The hash value.
    */
   public static raw md5Digest(Text data)
   {
      return md5Digest(data, (String) null);
   }
   
   /**
    * Hashes the specified data using RSA MD5 algorithm, and returns a 16-byte 
    * binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    * @param    key
    *           An optional key value to use in the hash operation.
    *           
    * @return   The hash value.
    */
   public static raw md5Digest(Text data, String key)
   {
      return md5Digest(data, new character(key));
   }
   
   /**
    * Hashes the specified data using RSA MD5 algorithm, and returns a 16-byte 
    * binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    * @param    key
    *           An optional key value to use in the hash operation.
    *           
    * @return   The hash value.
    */
   public static raw md5Digest(Text data, Text key)
   {
      // TODO:
      return null;
   }

   /**
    * Hashes the specified data using RSA MD5 algorithm, and returns a 16-byte 
    * binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    * @param    key
    *           An optional key value to use in the hash operation.
    *           
    * @return   The hash value.
    */
   public static raw md5Digest(Text data, BinaryData key)
   {
      // TODO: IMPLEMENT THIS
      return null;
   }
   
   /**
    * Hashes the specified data using RSA MD5 algorithm, and returns a 16-byte 
    * binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    *           
    * @return   The hash value.
    */
   public static raw md5Digest(BinaryData data)
   {
      return md5Digest(data, (String) null);
   }
   
   /**
    * Hashes the specified data using RSA MD5 algorithm, and returns a 16-byte 
    * binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    * @param    key
    *           An optional key value to use in the hash operation.
    *           
    * @return   The hash value.
    */
   public static raw md5Digest(BinaryData data, String key)
   {
      return md5Digest(data, new character(key));
   }
   
   /**
    * Hashes the specified data using RSA MD5 algorithm, and returns a 16-byte 
    * binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    * @param    key
    *           An optional key value to use in the hash operation.
    *           
    * @return   The hash value.
    */
   public static raw md5Digest(BinaryData data, Text key)
   {
      // TODO: IMPLEMENT THIS
      return null;
   }

   /**
    * Hashes the specified data using RSA MD5 algorithm, and returns a 16-byte 
    * binary message digest value in a {@link raw} instance.
    * 
    * @param    data
    *           The data to be hashed.
    * @param    key
    *           An optional key value to use in the hash operation.
    *           
    * @return   The hash value.
    */
   public static raw md5Digest(BinaryData data, BinaryData key)
   {
      // TODO: IMPLEMENT THIS
      return null;
   }

   /**
    * Returns the default userid used for authentication to databases that do not have _User
    * meta-table populated. Usually this is the username from the underlying OS, or empty string
    * if it cannot be detected for various causes.
    *
    * @return  the default userid used for authentication to databases.
    */
   public static String getDefaultUserid()
   {
      try
      {
         // extract the user name from the OS properties
         String defUserid = System.getProperty("user.name");

         // make sure is empty string instead of null
         return defUserid == null ? "" : defUserid;
      }
      catch (SecurityException e)
      {
         // the security manager does not allow extracting the information return empty string, too
         return "";
      }
   }
}
